--[[
Interface: 1.5.1.0 b6732

Copyright (C) GtX (Andy), 2020

Author: GtX | Andy
Date: 12.01.2020

Contact:
https://forum.giants-software.com
https://github.com/GtX-Andy

History:
V 1.0.0.0 @ 12.01.2020 - Release

Important:
No changes are to be made to this script without permission from GtX | Andy
An diesem Skript dürfen ohne Genehmigung von GtX | Andy keine Änderungen vorgenommen werden
]]


local manager
local loadError

local buildId = 1
local versionString = "0.0.0.0"

local modName = g_currentModName
local modDirectory = g_currentModDirectory

local function finalizeVehicleTypes(vehicleTypeManager)
    local numInserted = 0

    local specializationName = string.format("%s.livestockTrailerAddon", modName)
    local specializationObject = g_specializationManager:getSpecializationObjectByName(specializationName)

    if specializationObject ~= nil then
        for typeName, typeEntry in pairs(vehicleTypeManager.vehicleTypes) do
            if specializationObject.prerequisitesPresent(typeEntry.specializations) then
                g_vehicleTypeManager:addSpecialization(typeName, specializationName)

                numInserted = numInserted + 1
            end
        end
    end

    if manager ~= nil then
        manager.numSpecsInserted = numInserted
    end
end

local function livestockTrailerSaveToXMLFile_inj(vehicle, superFunc, xmlFile, key, usedModNames)
    if vehicle.spec_livestockTrailerAddon ~= nil then
        -- Avoids load errors if 'LivestockTrailerAddon' is removed with chickens in a trailer.
        if vehicle.spec_livestockTrailer.currentAnimalType == "CHICKEN" then
            key = string.format("%s.livestockTrailerAddon", key)
        end
    end

    superFunc(vehicle, xmlFile, key, usedModNames)
end

local function animalManagerLoadAnimals_inj(animalManager, xmlHandle, baseDirectory)
    if manager ~= nil then
        manager:loadAnimalTypesData(animalManager)
    end
end

local function animalLoadingTriggerOnCreate_inj(animalLoadingTrigger, superFunc, id)
    local trigger = AnimalLoadingTrigger:new(g_server ~= nil, g_client ~= nil)

    if trigger ~= nil then
        g_currentMission:addOnCreateLoadedObject(trigger)
        trigger:load(id)

        if manager ~= nil then
            if manager:getAddChickensToDealers() then
                manager:addChickensToTrigger(trigger, "AnimalLoadingTrigger")
            end
        end

        trigger:register(true)
    end
end

local function initLoadError(name)
    loadError = {}

    loadError.name = name
    loadError.modName = modName
    loadError.startUpdateTime = 2000
    loadError.versionString = versionString

    loadError.update = function(self, dt)
        if self.startUpdateTime > 0 then
            self.startUpdateTime = self.startUpdateTime - dt
        else
            local title = string.format("Livestock Trailer Addon - Version %s", self.versionString)
            local text = string.format(g_i18n:getText("LTA_loadError", self.modName), self.modName, self.name)

            if g_dedicatedServerInfo == nil then
                if not g_gui:getIsGuiVisible() then
                    g_gui:showYesNoDialog({
                        title = title,
                        text = text,
                        dialogType = DialogElement.TYPE_LOADING,
                        callback = self.openModHubLink,
                        target = nil,
                        yesText = g_i18n:getText("button_ok"),
                        noText = g_i18n:getText("button_modHubDownload")
                    })
                end
            else
                print("\n" .. text .. "\n    - https://farming-simulator.com/mods.php?lang=en&country=be&title=fs2019&filter=org&org_id=129652&page=0" .. "\n")
                self.openModHubLink(true)
            end
        end
    end

    loadError.openModHubLink = function(ignore)
        if ignore == false then
            local language = g_languageShort
            local link = "mods.php?lang=en&country=be&title=fs2019&filter=org&org_id=129652&page=0"
            if language == "de" or language == "fr" then
                link = "mods.php?lang=" .. language .. "&country=be&title=fs2019&filter=org&org_id=129652&page=0"
            end

            openWebFile(link, "")
        end

        removeModEventListener(loadError)
        loadError = nil
    end

    addModEventListener(loadError)

    g_storeManager.addModStoreItem = Utils.overwrittenFunction(g_storeManager.addModStoreItem, function (storeManager, superFunc, xmlFilename, baseDir, customEnvironment, isMod, isBundleItem, dlcTitle)
        if customEnvironment == modName then
            return
        end

        superFunc(storeManager, xmlFilename, baseDir, customEnvironment, isMod, isBundleItem, dlcTitle)
    end)
end

local function validate(name)
    local mod = g_modManager:getModByName(name)
    versionString = tostring(mod.version) or versionString

    if mod.modName == "FS19_LivestockTrailerAddon" or
        mod.modName == "FS19_LivestockTrailerAddon_update" then
        return mod.author:len() == 3, mod.author
    end

    return false, mod.author
end

local function deleteAddon()
    if manager ~= nil then
        manager:delete()

        manager = nil
        getfenv(0)["g_livestockTrailerAddon"] = nil
    end
end

local function loadMapFinished(mission, node, arguments, callAsyncCallback)
    if manager ~= nil then
        manager:loadGlobalVehiclesPlaceData()
        g_logManager:info("Livestock Trailer Addon - Version %s has been loaded into %d global vehicles successfully.", versionString, manager.numGlobalVehicles)

        if manager.invalidGlobalVehicles ~= nil and #manager.invalidGlobalVehicles > 0 then
            print("    - The following vehicles have been ignored due to their update version.")

            for i, data in pairs (manager.invalidGlobalVehicles) do
                print(string.format("        %d:  %s  -  %s", i, data.modName, data.reason))
            end

            manager.invalidGlobalVehicles = nil
            manager.invalidChickenVehicles = nil
        end
    end
end

local function initAddon()
    if g_livestockTrailerAddon ~= nil then
        g_logManager:error("Livestock Trailer Addon has already been loaded by '%s'! Please remove duplicate mod '%s'.", g_livestockTrailerAddon.modName, modName)
        return
    end

    local canLoad, author = validate(modName)
    if canLoad then
        source(modDirectory .. "scripts/LivestockTrailerAddonManager.lua")
        g_gui:loadProfiles(modDirectory .. "gui/guiProfiles.xml")

        g_specializationManager:addSpecialization("livestockTrailerAddon", "LivestockTrailerAddon", modDirectory .. "scripts/specializations/LivestockTrailerAddon.lua")

        manager = LivestockTrailerAddonManager:new(modDirectory, modName, buildId, versionString, g_animalManager, g_storeManager)
        manager:loadSettingsFromXML()

        getfenv(0)["g_livestockTrailerAddon"] = manager

        FSBaseMission.delete = Utils.appendedFunction(FSBaseMission.delete, deleteAddon)
        FSBaseMission.loadMapFinished = Utils.prependedFunction(FSBaseMission.loadMapFinished, loadMapFinished)
        VehicleTypeManager.finalizeVehicleTypes = Utils.prependedFunction(VehicleTypeManager.finalizeVehicleTypes, finalizeVehicleTypes)

        AnimalManager.loadAnimals = Utils.appendedFunction(AnimalManager.loadAnimals, animalManagerLoadAnimals_inj)
        AnimalLoadingTrigger.onCreate = Utils.overwrittenFunction(AnimalLoadingTrigger.onCreate, animalLoadingTriggerOnCreate_inj)
        LivestockTrailer.saveToXMLFile = Utils.overwrittenFunction(LivestockTrailer.saveToXMLFile, livestockTrailerSaveToXMLFile_inj)

        g_placeableTypeManager:addPlaceableType("chickenDealerPlaceable", "ChickenDealerPlaceable", modDirectory .. "scripts/placeables/ChickenDealerPlaceable.lua")
    else
        initLoadError(author)
    end
end

initAddon()
