--[[
Interface: 1.5.1.0 b6732

Copyright (C) GtX (Andy), 2020

Author: GtX | Andy
Date: 12.01.2020

Contact:
https://forum.giants-software.com
https://github.com/GtX-Andy

History:
V 1.0.0.0 @ 12.01.2020 - Release

Important:
No changes are to be made to this script without permission from GtX | Andy
An diesem Skript dürfen ohne Genehmigung von GtX | Andy keine Änderungen vorgenommen werden
]]


LivestockTrailerAddonManager = {}
local LivestockTrailerAddonManager_mt = Class(LivestockTrailerAddonManager)

function tableLength(table)
    local count = 0

    for _ in pairs(table) do
        count = count + 1
    end

    return count
end

function LivestockTrailerAddonManager:new(modDirectory, modName, buildId, versionString, animalManager, storeManager)
    self = {}
    setmetatable(self, LivestockTrailerAddonManager_mt)

    self.isServer = g_server ~= nil
    self.isClient = g_client ~= nil
    self.isDedi = g_dedicatedServerInfo ~= nil

    self.animalManager = animalManager
    self.storeManager = storeManager

    self.modName = modName
    self.modDirectory = modDirectory

    self.buildId = buildId
    self.versionString = versionString

    self.xmlFilePath = modDirectory .. "livestockTrailerAddon.xml"
    self.settingsFilePath = getUserProfileAppPath() .. "livestockTrailerAddon.xml"

    self.dealers = {}
    self.globalVehicles = {}

    self.invalidChickenVehicles = {}
    self.invalidGlobalVehicles = {}

    self.placeableDealers = {}
    self.numPlaceableDealers = 0

    self.numSpecsInserted = 0
    self.numGlobalVehicles = 0

    self.animalTypesDataLoaded = false
    self.animalTypeToFillTypeById = {}
    self.animalTypeToFirstFillType = {}

    self.canAddChickensToDealers = true
    self.playTransportSounds = not self.isDedi
    self.showAnimalFillLevel = not self.isDedi

    return self
end

function LivestockTrailerAddonManager:delete()
    -- Debug cleanup
end

function LivestockTrailerAddonManager:loadAnimalTypesData(animalManager)
    if self.animalTypesDataLoaded then
        return
    end

    self:verifyChickenCanBeBought(animalManager, false)

    for _, animalType in pairs (animalManager.animals) do
        local animalTypeName = animalType.type
        self.animalTypeToFillTypeById[animalTypeName] = {}

        for _, subType in pairs (animalType.subTypes) do
            local subTypeId = subType.subTypeId

            if subTypeId == 1 then
                self.animalTypeToFirstFillType[animalTypeName] = subType.fillType
            end

            self.animalTypeToFillTypeById[animalTypeName][subTypeId] = subType.fillType
        end
    end

    self.animalTypesDataLoaded = true
end

function LivestockTrailerAddonManager:loadGlobalVehiclesPlaceData()
    local xmlFile = loadXMLFile("VehiclesDataXML", self.xmlFilePath)
    if xmlFile ~= nil and xmlFile ~= 0 then
        local filename = getXMLString(xmlFile, "livestockTrailerAddon.globalAnimalPlaces#filename")
        local insertCapacitySpecType = self.storeManager.nameToSpecType ~= nil
        local isMod, invalidChickenVehicle = false, nil

        if filename ~= nil then
            local i = 0
            while true do
                local key = string.format("livestockTrailerAddon.globalAnimalPlaces.vehicle(%d)", i)
                if not hasXMLProperty(xmlFile, key) then
                    break
                end

                local configFileName = getXMLString(xmlFile, key .. "#xmlFilename")
                if configFileName ~= nil then
                    local directoryPrefix = getXMLString(xmlFile, key .. "#directoryPrefix")

                    if directoryPrefix ~= nil then
                        if directoryPrefix == "$pdlcdir$" or directoryPrefix == "$moddir$" then
                            configFileName = NetworkUtil.convertFromNetworkFilename(directoryPrefix .. configFileName)
                            isMod = true

                            local modName, _ = Utils.getModNameAndBaseDirectory(configFileName)
                            if modName ~= nil then
                                if g_modIsLoaded[modName] then
                                    invalidChickenVehicle = self.invalidChickenVehicles[modName]
                                else
                                    configFileName = ""
                                end
                            else
                                configFileName = ""
                            end
                        end
                    end

                    if configFileName ~= "" then
                        configFileName = configFileName:lower()

                        if self.globalVehicles[configFileName] == nil then
                            local version = Utils.getNoNil(getXMLString(xmlFile, key .. "#version"), "1.0.0.0")

                            if invalidChickenVehicle == nil then
                                self.globalVehicles[configFileName] = {
                                    places = {},
                                    isMod = isMod,
                                    version = version
                                }

                                local j = 0
                                while true do
                                    local animalTypeKey = string.format("%s.animal(%d)", key, j)
                                    if not hasXMLProperty(xmlFile, animalTypeKey) then
                                        break
                                    end

                                    local animalTypeName = getXMLString(xmlFile, animalTypeKey .. "#type")
                                    if animalTypeName ~= nil then
                                        animalTypeName = animalTypeName:upper()

                                        local place = {}

                                        place.canLoad = true
                                        place.node = Utils.getNoNil(getXMLString(xmlFile, animalTypeKey .. "#node"), "0")
                                        place.backupLinkNode = Utils.getNoNil(getXMLString(xmlFile, animalTypeKey .. "#backupLinkNode"), "0>")
                                        place.numSlots = getXMLInt(xmlFile, animalTypeKey .. "#numSlots") or 0
                                        place.i3dFilename = filename
                                        place.baseDirectory = self.modDirectory

                                        self.globalVehicles[configFileName].places[animalTypeName] = place

                                        if insertCapacitySpecType and place.numSlots > 0 then
                                            -- numAnimalsCow, numAnimalsPig, numAnimalsSheep, numAnimalsHorse, numAnimalsChicken
                                            local capacitySpecType = getXMLString(xmlFile, animalTypeKey .. "#capacitySpecType")

                                            if capacitySpecType ~= nil and self.storeManager.nameToSpecType[capacitySpecType] ~= nil then
                                                local item = self.storeManager.xmlFilenameToItem[configFileName]

                                                if item ~= nil and item.specs ~= nil then
                                                    -- Only if mod has not built in support itself.
                                                    if item.specs[capacitySpecType] == nil then
                                                        item.specs[capacitySpecType] = place.numSlots
                                                    end
                                                end
                                            end
                                        end
                                    end

                                    j = j + 1
                                end

                                self.numGlobalVehicles = self.numGlobalVehicles + 1
                            else
                                table.insert(self.invalidGlobalVehicles, invalidChickenVehicle)
                            end
                        else
                            print(string.format("  Debug: [Livestock Trailer Addon]  Ignoring duplicate global vehicle entry at '%s'!  (%s)", key, self.xmlFilePath))
                        end
                    end
                end

                i = i + 1
            end
        end

        delete(xmlFile)
    end
end

function LivestockTrailerAddonManager:setInvalidGlobalVehicle(customEnvironment, data)
    if customEnvironment ~= nil then
        self.invalidChickenVehicles[customEnvironment] = data
    end
end

function LivestockTrailerAddonManager:verifyChickenCanBeBought(animalManager, printInfo)
    local numCorrected = 0

    if animalManager == nil then
        animalManager = self.animalManager
    end

    if animalManager.typeToAnimal.CHICKEN ~= nil then
        for _, animalSubType in pairs (animalManager.typeToAnimal.CHICKEN.subTypes) do
            if not animalSubType.storeInfo.canBeBought then
                animalSubType.storeInfo.canBeBought = true
                numCorrected = numCorrected + 1
            end
        end
    end

    if numCorrected > 0 and printInfo then
        g_logManager:info("[Livestock Trailer Addon] Chicken store info was changed by another mod, this has been corrected.")
    end

    return numCorrected
end

function LivestockTrailerAddonManager:addPlaceableDealer(dealer)
    if self.placeableDealers[dealer] == nil then
        self.placeableDealers[dealer] = dealer
        self.numPlaceableDealers = tableLength(self.placeableDealers)

        return true
    end

    return false
end

function LivestockTrailerAddonManager:removePlaceableDealer(dealer)
    if self.placeableDealers[dealer] ~= nil then
        self.placeableDealers[dealer] = nil
        self.numPlaceableDealers = tableLength(self.placeableDealers)

        return true
    end

    return false
end

function LivestockTrailerAddonManager:getNumPlaceableDealers()
    return self.numPlaceableDealers
end

function LivestockTrailerAddonManager:loadSettingsFromXML()
    if self.isDedi then
        return
    end

    if fileExists(self.settingsFilePath) then
        local xmlFile = loadXMLFile("LivestockTrailerAddonXML", self.settingsFilePath)

        if xmlFile ~= nil and xmlFile ~= 0 then
            if g_careerScreen ~= nil and g_careerScreen.isMultiplayer then
                self.canAddChickensToDealers = true -- Always enable in MP so there is no issues as XML is for each player.
            else
                self.canAddChickensToDealers = Utils.getNoNil(getXMLBool(xmlFile, "livestockTrailerAddon.addChickensToDealers"), true)
            end
            self.playTransportSounds = Utils.getNoNil(getXMLBool(xmlFile, "livestockTrailerAddon.useTransportSounds"), true)
            self.showAnimalFillLevel = Utils.getNoNil(getXMLBool(xmlFile, "livestockTrailerAddon.showVehicleAnimalFillLevel"), true)

            delete(xmlFile)
        end
    else
        local xmlFile = io.open (self.settingsFilePath, "w")
        if xmlFile ~= nil and xmlFile ~= 0 then
            local xmlText = '<?xml version="1.0" encoding="utf-8" standalone="no" ?>\n\n' ..
            '<livestockTrailerAddon build="' .. tostring(self.buildId) .. '" version="' .. self.versionString .. '">\n' ..
            '    <useTransportSounds>true</useTransportSounds> <!-- The animals loaded in your trailer will randomly make sounds. -->\n' ..
            '    <addChickensToDealers>true</addChickensToDealers> <!-- Chickens will be added to all dealers built into the map with more than one animal type. Feature is always enabled in MP! -->\n' ..
            '    <showVehicleAnimalFillLevel>true</showVehicleAnimalFillLevel> <!-- Shows the fill percentage and animal type in lower left of screen. -->\n' ..
            '</livestockTrailerAddon>'

            xmlFile:write(xmlText)
            xmlFile:close()
        end
    end
end

function LivestockTrailerAddonManager:getValidDummyFilename(fillTypeIndex)
    if fillTypeIndex == FillType.CHICKEN_TYPE_ROOSTER then
        return self.modDirectory .. "character/roosterLoaded.i3d"
    end

    return self.modDirectory .. "character/henLoaded.i3d"
end

function LivestockTrailerAddonManager:getIsGlobalVehicle(configFileName)
    return self.globalVehicles[configFileName:lower()] ~= nil
end

function LivestockTrailerAddonManager:getGlobalVehicleData(configFileName)
    return self.globalVehicles[configFileName:lower()]
end

function LivestockTrailerAddonManager:getGlobalVehicleSlotData(configFileName, animalTypeName)
    configFileName = configFileName:lower()

    if self.globalVehicles[configFileName] ~= nil then
        return self.globalVehicles[configFileName].places[animalTypeName]
    end

    return nil
end

function LivestockTrailerAddonManager:getAddChickensToDealers(maxPlaceables)
    if self.canAddChickensToDealers and self.numPlaceableDealers < Utils.getNoNil(maxPlaceables, math.huge) then
        return self.animalManager:getAnimalsByType("CHICKEN") ~= nil
    end

    return false
end

function LivestockTrailerAddonManager:addChickensToDealers(className, maxPlaceables)
    local numAdded = 0

    if self:getAddChickensToDealers(maxPlaceables) then
        if g_currentMission ~= nil and not g_currentMission.isExitingGame then
            if g_currentMission.onCreateLoadedObjects ~= nil then
                className = className or "AnimalLoadingTrigger"

                for _, object in pairs (g_currentMission.onCreateLoadedObjects) do
                    if self:addChickensToTrigger(object, className) then
                        numAdded = numAdded + 1
                    end
                end
            end
        end
    end

    return numAdded
end

function LivestockTrailerAddonManager:removeChickensFromDealers(className)
    local numRemoved = 0

    if self:getAddChickensToDealers() then
        if g_currentMission ~= nil and not g_currentMission.isExitingGame then
            if g_currentMission.onCreateLoadedObjects ~= nil then
                className = className or "AnimalLoadingTrigger"

                for _, object in pairs (g_currentMission.onCreateLoadedObjects) do
                    if self:removeChickensFromTrigger(object, className) then
                        numRemoved = numRemoved + 1
                    end
                end
            end
        end
    end

    return numRemoved
end

function LivestockTrailerAddonManager:addChickensToTrigger(trigger, className)
    if trigger.isDealer and trigger.className == className then
        local chickenAnimalType = self.animalManager:getAnimalsByType("CHICKEN")

        if chickenAnimalType ~= nil then
            -- Allow map makers to block the addition of chicken using attribue 'allowLivestockTrailerAddon'.
            trigger.allowLivestockTrailerAddon = Utils.getNoNil(getUserAttribute(trigger.triggerNode, "allowLivestockTrailerAddon"), true)

            if trigger.allowLivestockTrailerAddon then
                -- I only insert into trigger with more than one animal type. Chickens are no good at a cattle dealer.
                if trigger.animalTypes == nil or (tableLength(trigger.animalTypes) == 1) then
                    trigger.allowLivestockTrailerAddon = false
                    return false
                end

                for _, animalType in pairs (trigger.animalTypes) do
                    if animalType == chickenAnimalType then
                        trigger.allowLivestockTrailerAddon = false
                        return false
                    end
                end

                trigger.hasChickens = true
                self.dealers[trigger] = trigger

                table.insert(trigger.animalTypes, chickenAnimalType)

                return true
            end
        end
    end

    return false
end

function LivestockTrailerAddonManager:removeChickensFromTrigger(trigger, className)
    if trigger.isDealer and trigger.className == className then
        if trigger.animalTypes == nil or (tableLength(trigger.animalTypes) == 1) then
            return false
        end

        local chickenAnimalType = self.animalManager:getAnimalsByType("CHICKEN")

        if chickenAnimalType ~= nil then
            for i, animalType in pairs (trigger.animalTypes) do
                if animalType == chickenAnimalType then
                    trigger.hasChickens = false
                    table.remove(trigger.animalTypes, i)
                    self.dealers[trigger] = nil

                    return true
                end
            end
        end
    end

    return false
end

function LivestockTrailerAddonManager:getFillTypeFromAnimalType(animalType, subTypeId)
    if self.animalTypeToFillTypeById[animalType] ~= nil then
        return self.animalTypeToFillTypeById[animalType][subTypeId]
    end

    return nil
end

function LivestockTrailerAddonManager:getFirstFillTypeFromAnimalType(animalType)
    return self.animalTypeToFirstFillType[animalType]
end
