--[[
Interface: 1.5.1.0 b6732

Copyright (C) GtX (Andy), 2020

Author: GtX | Andy
Date: 19.01.2020

Contact:
https://forum.giants-software.com
https://github.com/GtX-Andy

History:
V 1.0.0.0 @ 19.01.2020 - Release Version

Important:
No changes are to be made to this script without permission from GtX | Andy
An diesem Skript dürfen ohne Genehmigung von GtX | Andy keine Änderungen vorgenommen werden
]]


ChickenDealerPlaceable = {}

ChickenDealerPlaceable.TIMER = 3000

ChickenDealerPlaceable.MAX_SAMPLES = 5
ChickenDealerPlaceable.MAX_CLIPS = 10
ChickenDealerPlaceable.MAX_PLACEABLES = 1

ChickenDealerPlaceable.COLOURS = {
    WHITE = {2, 1},
    BLACK = {1, 1},
    BROWN = {2, 2}
}

local ChickenDealerPlaceable_mt = Class(ChickenDealerPlaceable, Placeable)
InitObjectClass(ChickenDealerPlaceable, "ChickenDealerPlaceable")

function ChickenDealerPlaceable:new(isServer, isClient, customMt)
    local self = Placeable:new(isServer, isClient, customMt or ChickenDealerPlaceable_mt)

    registerObjectClassName(self, "ChickenDealerPlaceable")

    self.animalLoadingTrigger = nil
    self.buyingStation = nil

    self.animationsActive = true
    self.chickenAnimations = {}
    self.animationTimer = ChickenDealerPlaceable.TIMER

    self.vehiclesInRange = {}
    self.numVehiclesInRange = 0

    self.chickenSamples = {}
    self.numChickenSamples = 0
    self.chickenSampleDelay = 0

    return self
end

function ChickenDealerPlaceable:load(xmlFilename, x,y,z, rx,ry,rz, initRandom)
    -- Correct the 'fullName' using customEnvironment.
    self.loadHotspotFromXML = Utils.overwrittenFunction(self.loadHotspotFromXML, function(placeable, superFunc, xmlFile, key)
        local hotspot = superFunc(placeable, xmlFile, key)

        if hotspot ~= nil then
            local fullViewName = g_i18n:convertText(Utils.getNoNil(getXMLString(xmlFile, key .. "#fullName"), ""), placeable.customEnvironment)
            if fullViewName ~= "" then
                hotspot.fullViewName = fullViewName
            end
        end

        return hotspot
    end)

    if not ChickenDealerPlaceable:superClass().load(self, xmlFilename, x,y,z, rx,ry,rz, initRandom) then
        return false
    end

    local xmlFile = loadXMLFile("TempXML", xmlFilename)

    local triggerNode = I3DUtil.indexToObject(self.nodeId, getXMLString(xmlFile, "placeable.chickenDealer.animalTrigger#node"))
    if triggerNode ~= nil then
        local trigger = AnimalLoadingTrigger:new(self.isServer, self.isClient)

        if trigger ~= nil then
            trigger.isEnabled = true
            trigger.isDealer = true

            trigger.triggerNode = triggerNode
            addTrigger(triggerNode, "triggerCallback", trigger)

            trigger.customEnvironment = self.customEnvironment
            trigger.title = g_i18n:getText(Utils.getNoNil(getXMLString(xmlFile, "placeable.chickenDealer#title"), "animals_dealer"), trigger.customEnvironment)

            local animalType = g_animalManager:getAnimalsByType("CHICKEN")
            if animalType ~= nil then
                trigger.animalTypes = {animalType}
            else
                g_logManager:xmlWarning(xmlFilename, "Animal type 'CHICKEN' is not available on this map, dealer disabled!")
            end

            trigger.mapHotspot = nil

            self.animalLoadingTrigger = trigger
        end
    end

    if hasXMLProperty(xmlFile, "placeable.chickenDealer.buyingStation") then
        self.buyingStation = BuyingStation:new(self.isServer, self.isClient)
        self.buyingStation:load(self.nodeId, xmlFile, "placeable.chickenDealer.buyingStation", self.customEnvironment)
        self.buyingStation.owningPlaceable = self
    end

    if self.isClient then
        self.animationTrigger = I3DUtil.indexToObject(self.nodeId, getXMLString(xmlFile, "placeable.chickenDealer.animationTrigger#node"))

        if self.animationTrigger ~= nil then
            local i = 0
            while true do
                local key = string.format("placeable.chickenDealer.animationClips.animationClip(%d)", i)
                if not hasXMLProperty(xmlFile, key) then
                    break
                end

                local rootNode = I3DUtil.indexToObject(self.nodeId, getXMLString(xmlFile, key .. "#rootNode"))
                if rootNode ~= nil then
                    local animCharacterSet = getAnimCharacterSet(rootNode)

                    if animCharacterSet ~= nil then
                        local clip = getAnimClipIndex(animCharacterSet, getXMLString(xmlFile, key .. "#clip"))

                        assignAnimTrackClip(animCharacterSet, 0, clip)
                        setAnimTrackSpeedScale(animCharacterSet, 0, 1)
                        setAnimTrackLoopState(animCharacterSet, 0, true)

                        local animation = {}
                        animation.rootNode = rootNode
                        animation.characterSet = animCharacterSet
                        animation.duration = getAnimClipDuration(animCharacterSet, clip)

                        local animalMesh = I3DUtil.indexToObject(self.nodeId, getXMLString(xmlFile, key .. "#shaderNode"))
                        if animalMesh ~= nil then
                            local colourName = Utils.getNoNil(getXMLString(xmlFile, key .. "#colorName"), "WHITE")
                            local colour = ChickenDealerPlaceable.COLOURS[colourName:upper()]

                            if colour == nil then
                                colour = ChickenDealerPlaceable.COLOURS["WHITE"]
                            end

                            local x, y, z, w = getShaderParameter(animalMesh, "RDT")
                            setShaderParameter(animalMesh, "RDT", x, math.random(0, 10) * 0.1, z, w, false)

                            local x, y, _, _ = getShaderParameter(animalMesh, "atlasInvSizeAndOffsetUV")
                            local tileU = colour[1] / (1 / x)
                            local tileV = colour[2] / (1 / y)

                            setShaderParameter(animalMesh, "atlasInvSizeAndOffsetUV", x, y, tileU, tileV, false)
                        end

                        setAnimTrackTime(animation.characterSet, 0, math.random(0, animation.duration), true)
                        enableAnimTrack(animation.characterSet, 0)

                        table.insert(self.chickenAnimations, animation)

                        if #self.chickenAnimations == ChickenDealerPlaceable.MAX_CLIPS then
                            break
                        end
                    end
                end

                i = i + 1
            end

            self.samplesStartTime = Utils.getNoNil(getXMLFloat(xmlFile, "placeable.chickenDealer.samples#startTime"), 0)
            self.samplesEndTime = Utils.getNoNil(getXMLFloat(xmlFile, "placeable.chickenDealer.samples#endTime"), 24)

            i = 0
            while true do
                local sampleKey = string.format("sample(%d)", i)
                if not hasXMLProperty(xmlFile, "placeable.chickenDealer.samples." .. sampleKey) then
                    break
                end

                local sample = g_soundManager:loadSampleFromXML(xmlFile, "placeable.chickenDealer.samples", sampleKey, self.baseDirectory, self.nodeId, 1, AudioGroup.ENVIRONMENT, nil, nil)
                if sample ~= nil then
                    self.numChickenSamples = self.numChickenSamples + 1
                    self.chickenSamples[self.numChickenSamples] = sample

                    if self.numChickenSamples == ChickenDealerPlaceable.MAX_SAMPLES then
                        break
                    end
                end

                i = i + 1
            end

            addTrigger(self.animationTrigger, "animationTriggerCallback", self)
        end
    end

    delete(xmlFile)
    xmlFile = nil

    return true
end

function ChickenDealerPlaceable:finalizePlacement()
    ChickenDealerPlaceable:superClass().finalizePlacement(self)

    if g_livestockTrailerAddon ~= nil then
        g_livestockTrailerAddon:addPlaceableDealer(self)
        g_livestockTrailerAddon:removeChickensFromDealers("AnimalLoadingTrigger")
    end

    if self.animalLoadingTrigger ~= nil then
        self.animalLoadingTrigger:register(true)
    end

    if self.buyingStation ~= nil then
        self.buyingStation:register(true)
    end

    -- This is not updated on placement only when loading from save game. No good if using store hours so push update to check on placement.
    if self.isServer then
        if self.animatedObjects ~= nil then
            for _, animatedObject in pairs(self.animatedObjects) do
                AnimatedObject.hourChanged(animatedObject)
            end
        end
    end
end

function ChickenDealerPlaceable:delete()
    if g_livestockTrailerAddon ~= nil then
        if g_livestockTrailerAddon:removePlaceableDealer(self) then
            g_livestockTrailerAddon:addChickensToDealers("AnimalLoadingTrigger", 1)
        end
    end

    if self.animalLoadingTrigger ~= nil then
        self.animalLoadingTrigger:delete()
        self.animalLoadingTrigger = nil
    end

    if self.buyingStation ~= nil then
        self.buyingStation:delete()
    end

    if self.isClient then
        if self.animationTrigger ~= nil then
            removeTrigger(self.animationTrigger)
            self.animationTrigger = nil
        end

        if self.numChickenSamples > 0 then
            g_soundManager:deleteSamples(self.chickenSamples)
            self.numChickenSamples = 0
        end
    end

    unregisterObjectClassName(self)
    ChickenDealerPlaceable:superClass().delete(self)
end

function ChickenDealerPlaceable:readStream(streamId, connection)
    ChickenDealerPlaceable:superClass().readStream(self, streamId, connection)

    if connection:getIsServer() then
        local buyingStationId = NetworkUtil.readNodeObjectId(streamId)

        self.buyingStation:readStream(streamId, connection)
        g_client:finishRegisterObject(self.buyingStation, buyingStationId)
    end
end

function ChickenDealerPlaceable:writeStream(streamId, connection)
    ChickenDealerPlaceable:superClass().writeStream(self, streamId, connection)

    if not connection:getIsServer() then
        NetworkUtil.writeNodeObjectId(streamId, NetworkUtil.getObjectId(self.buyingStation))
        self.buyingStation:writeStream(streamId, connection)
        g_server:registerObjectInStream(connection, self.buyingStation)
    end
end

function ChickenDealerPlaceable:collectPickObjects(node)
    local foundNode = false

    if not foundNode then
        for _, loadTrigger in ipairs(self.buyingStation.loadTriggers) do
            if node == loadTrigger.triggerNode then
                foundNode = true

                break
            end
        end
    end

    if not foundNode then
        ChickenDealerPlaceable:superClass().collectPickObjects(self, node)
    end
end

function ChickenDealerPlaceable:getCanBePlacedAt(x, y, z, distance, farmId)
    local canBePlaced = ChickenDealerPlaceable:superClass().getCanBePlacedAt(self, x, y, z, distance, farmId)
    return canBePlaced and self:getUnderPlacementLimit()
end

function ChickenDealerPlaceable:canBuy()
    local canBuy = ChickenDealerPlaceable:superClass().canBuy(self)

    if canBuy then
        return self:getUnderPlacementLimit(), g_i18n:getText("warning_tooManyPlaceables")
    end

    return canBuy, g_i18n:getText("warning_placeable_error_cannotBeBought")
end

function ChickenDealerPlaceable:getUnderPlacementLimit()
    if g_livestockTrailerAddon ~= nil then
        return g_livestockTrailerAddon:getNumPlaceableDealers() < ChickenDealerPlaceable.MAX_PLACEABLES
    end

    return true
end

function ChickenDealerPlaceable:updateTick(dt)
    ChickenDealerPlaceable:superClass().updateTick(self, dt)

    if self.isClient then
        if self:getCanInteract() then
            self.animationTimer = ChickenDealerPlaceable.TIMER
            self:setAnimationAndSamples(true)

            if self:getCanPlayChickenSounds(g_currentMission.environment.currentHour) then
                if self.chickenSampleDelay < g_currentMission.time then
                    local sample = self.chickenSamples[math.random(1, self.numChickenSamples)]
                    g_soundManager:playSample(sample)

                    self.chickenSampleDelay = g_currentMission.time + math.max(math.random(4000, 8000), sample.duration)
                end
            end
        else
            self.animationTimer = self.animationTimer - dt

            if self.animationTimer > 0 then
                self:raiseActive()
            else
                self:setAnimationAndSamples(false)
            end
        end
    end
end

function ChickenDealerPlaceable:getCanPlayChickenSounds(currentHour)
    if self.numChickenSamples > 0 and currentHour ~= nil then
        return currentHour >= self.samplesStartTime and currentHour < self.samplesEndTime
    end

    return false
end

function ChickenDealerPlaceable:setAnimationAndSamples(active)
    if self.animationsActive ~= active then
        self.animationsActive = active

        for _, animation in pairs (self.chickenAnimations) do
            if self.animationsActive then
                setAnimTrackTime(animation.characterSet, 0, math.random(0, animation.duration))
                enableAnimTrack(animation.characterSet, 0)
            else
                disableAnimTrack(animation.characterSet, 0)
            end
        end

        if not self.animationsActive then
            for _, sample in pairs(self.chickenSamples) do
                g_soundManager:stopSample(sample)
            end
        end

        self.chickenSampleDelay = g_currentMission.time
    end
end

function ChickenDealerPlaceable:getCanInteract()
    if g_currentMission.controlPlayer then
        if self.playerInRange then
            self:raiseActive()
            return true
        end
    else
        if self.numVehiclesInRange > 0 then
            self:raiseActive()

            for vehicle, _ in pairs(self.vehiclesInRange) do
                if vehicle:getIsActiveForInput(true) then
                    return true
                end
            end
        end
    end

    return false
end

function ChickenDealerPlaceable:animationTriggerCallback(triggerId, otherId, onEnter, onLeave, onStay, otherShapeId)
    if onEnter or onLeave then
        if g_currentMission.player ~= nil and otherId == g_currentMission.player.rootNode then
            if onEnter then
                self.playerInRange = true
            else
                self.playerInRange = false
            end
        else
            local vehicle = g_currentMission.nodeToObject[otherShapeId]
            if vehicle ~= nil then
                if onEnter then
                    if self.vehiclesInRange[vehicle] == nil then
                        self.vehiclesInRange[vehicle] = true
                        self.numVehiclesInRange = self.numVehiclesInRange + 1
                    end
                else
                    if self.vehiclesInRange[vehicle] then
                        self.vehiclesInRange[vehicle] = nil
                        self.numVehiclesInRange = math.max(self.numVehiclesInRange - 1, 0)
                    end
                end
            end
        end

        self:raiseActive()
    end
end
