--[[
Interface: 1.5.1.0 b6732

Copyright (C) GtX (Andy), 2020

Author: GtX | Andy
Date: 12.01.2020

Contact:
https://forum.giants-software.com
https://github.com/GtX-Andy

History:
V 1.0.0.0 @ 12.01.2020 - Release

Important:
No changes are to be made to this script without permission from GtX | Andy
Use of this specialization within other mods is strictly prohibited!

An diesem Skript dürfen ohne Genehmigung von GtX | Andy keine Änderungen vorgenommen werden
Die Verwendung dieser Spezialisierung innerhalb anderer Mods ist strengstens untersagt!
]]


LivestockTrailerAddon = {}

LivestockTrailerAddon.MOD_NAME = g_currentModName
LivestockTrailerAddon.MAXIMUM_ANIMAL_TYPE_SOUNDS = 6

LivestockTrailerAddon.INVALID_DUMMY_NAMES = {
    ["dataS2/character/chicken/hen.i3d"] = true,
    ["dataS2/character/chicken/rooster.i3d"] = true
}

function LivestockTrailerAddon.initSpecialization()
    g_storeManager:addSpecType("numAnimalsChicken", "shopListAttributeIconChicken", LivestockTrailerAddon.loadSpecValueNumberAnimalsChicken, LivestockTrailerAddon.getSpecValueNumberAnimalsChicken)
end

function LivestockTrailerAddon.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(LivestockTrailer, specializations) and g_livestockTrailerAddon ~= nil
end

function LivestockTrailerAddon.registerFunctions(vehicleType)
    SpecializationUtil.registerFunction(vehicleType, "loadAnimalTypeFromXML",     LivestockTrailerAddon.loadAnimalTypeFromXML)
    SpecializationUtil.registerFunction(vehicleType, "loadTransportPlaces",       LivestockTrailerAddon.loadTransportPlaces)
    SpecializationUtil.registerFunction(vehicleType, "loadTransportSounds",       LivestockTrailerAddon.loadTransportSounds)
    SpecializationUtil.registerFunction(vehicleType, "getGlobalSlotLinkNode",     LivestockTrailerAddon.getGlobalSlotLinkNode)
    SpecializationUtil.registerFunction(vehicleType, "playTransportSounds",       LivestockTrailerAddon.playTransportSounds)
    SpecializationUtil.registerFunction(vehicleType, "getCanPlayTransportSounds", LivestockTrailerAddon.getCanPlayTransportSounds)
    SpecializationUtil.registerFunction(vehicleType, "getValidDummyFilename",     LivestockTrailerAddon.getValidDummyFilename)
end

function LivestockTrailerAddon.registerOverwrittenFunctions(vehicleType)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "getFillLevelInformation",  LivestockTrailerAddon.getFillLevelInformation)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "addAnimal",                LivestockTrailerAddon.addAnimal)
end

function LivestockTrailerAddon.registerEventListeners(vehicleType)
    SpecializationUtil.registerEventListener(vehicleType, "onLoad",              LivestockTrailerAddon)
    SpecializationUtil.registerEventListener(vehicleType, "onLoadFinished",      LivestockTrailerAddon)
    SpecializationUtil.registerEventListener(vehicleType, "onDelete",            LivestockTrailerAddon)
    SpecializationUtil.registerEventListener(vehicleType, "onUpdate",            LivestockTrailerAddon)
    SpecializationUtil.registerEventListener(vehicleType, "onReadUpdateStream",  LivestockTrailerAddon)
    SpecializationUtil.registerEventListener(vehicleType, "onWriteUpdateStream", LivestockTrailerAddon)
end

function LivestockTrailerAddon:onLoad(savegame)
    self.spec_livestockTrailerAddon = self[string.format("spec_%s.livestockTrailerAddon", LivestockTrailerAddon.MOD_NAME)]

    local spec = self.spec_livestockTrailerAddon

    spec.addonLoadedChickens = false

    spec.sampleIndexOne = 0
    spec.sampleIndexTwo = 0

    spec.soundDelayTime = 0
    spec.soundDelayInterval = 0

    spec.lastAnimalType = nil

    spec.animalTypeSamples = {}
    spec.animalTypePlaces = {}

    if g_livestockTrailerAddon ~= nil then
        -- If a mod vehicle and the XML property exists then load this first.
        if self.customEnvironment ~= nil and hasXMLProperty(self.xmlFile, "vehicle.livestockTrailerAddon") then
            self:loadAnimalTypeFromXML(self.xmlFile, "vehicle.livestockTrailerAddon", self.configFileName, self.baseDirectory, false)
        end

        local configFileName = g_livestockTrailerAddon.xmlFilePath
        local xmlFile = loadXMLFile("VehiclesDataXML", configFileName)

        if xmlFile ~= nil and xmlFile ~= 0 then
            self:loadAnimalTypeFromXML(xmlFile, "livestockTrailerAddon.animalTypeSamples", configFileName, g_livestockTrailerAddon.modDirectory, true)
            delete(xmlFile)
        else
            g_logManager:warning("[%s] Failed to load '%s', some vehicle features may not work!", LivestockTrailerAddon.MOD_NAME, configFileName)
        end

        if spec.addonLoadedChickens and self.ownerFarmId ~= nil then
            g_livestockTrailerAddon:verifyChickenCanBeBought(nil, true)
        end
    else
        g_logManager:warning("[%s] Some vehicle features failed to load and may not work correctly!", LivestockTrailerAddon.MOD_NAME)
    end

    spec.dirtyFlag = self:getNextDirtyFlag()
end

function LivestockTrailerAddon:onLoadFinished(savegame)
    local livestockTrailerSpec = self.spec_livestockTrailer

    if savegame ~= nil and not savegame.resetVehicles then
        -- Only if not loaded by base spec already.
        if livestockTrailerSpec.currentAnimalType == nil then
            local key = string.format("%s.livestockTrailer.livestockTrailerAddon", savegame.key)
            local i = 0
            local xmlFile = savegame.xmlFile

            while true do
                local slotKey = string.format("%s.animal(%d)", key, i)

                if not hasXMLProperty(xmlFile, slotKey) then
                    break
                end

                local animal = Animal.createFromXMLFile(xmlFile, slotKey, self.isServer, self.isClient, nil)

                animal:register()
                self:addAnimal(animal)

                i = i + 1
            end
        end
    end
end

function LivestockTrailerAddon:loadAnimalTypeFromXML(xmlFile, baseKey, configFileName, baseDirectory, globalLoad)
    local i = 0
    while true do
        local key = string.format("%s.animal(%d)", baseKey, i)
        if not hasXMLProperty(xmlFile, key) then
            break
        end

        local animalTypeName = getXMLString(xmlFile, key .. "#type")

        if animalTypeName ~= nil then
            animalTypeName = animalTypeName:upper()

            if g_livestockTrailerAddon:getFirstFillTypeFromAnimalType(animalTypeName) ~= nil then
                local animalType = g_animalManager:getAnimalsByType(animalTypeName)

                self:loadTransportPlaces(xmlFile, key, animalType, configFileName, globalLoad)
                self:loadTransportSounds(xmlFile, key .. ".samples", animalType, configFileName, baseDirectory, globalLoad)
            end
        else
            g_logManager:xmlWarning(configFileName, "No name given for '%s'", key)
        end

        i = i + 1
    end
end

function LivestockTrailerAddon:loadTransportPlaces(xmlFile, key, animalType, configFileName, globalLoad)
    local spec = self.spec_livestockTrailerAddon
    local livestockTrailerSpec = self.spec_livestockTrailer

    local node
    local numSlots = 0

    if livestockTrailerSpec.animalTypeToPlaces[animalType.type] ~= nil then
        if globalLoad and self.ownerFarmId ~= nil then
            local slotData = g_livestockTrailerAddon:getGlobalVehicleSlotData(self.configFileName, animalType.type)

            if slotData ~= nil and slotData.canLoad then
                slotData.canLoad = false
                g_logManager:info("[Livestock Trailer Addon]  Unknown version of '%s' loaded!  %s data has been changed and is being loaded directly from mod.", self.customEnvironment, animalType.type)
            end
        end

        return
    end

    if globalLoad then
        local slotData = g_livestockTrailerAddon:getGlobalVehicleSlotData(self.configFileName, animalType.type)

        if slotData ~= nil and slotData.canLoad then
            local i3dNode = g_i3DManager:loadSharedI3DFile(slotData.i3dFilename, slotData.baseDirectory, false, false, false)

            if i3dNode ~= 0 then
                local linkNode = self:getGlobalSlotLinkNode(slotData)

                if linkNode ~= nil then
                    numSlots = math.abs(slotData.numSlots or 0)
                    node = I3DUtil.indexToObject(i3dNode, slotData.node)

                    if node ~= nil then
                        local placeData = {}

                        placeData.node = node
                        placeData.linkNode = linkNode
                        placeData.animalType = animalType.type
                        placeData.filename = slotData.i3dFilename
                        placeData.baseDirectory = slotData.baseDirectory

                        link(placeData.linkNode, placeData.node)

                        table.insert(spec.animalTypePlaces, placeData)
                    end
                end
            end
        end
    else
        node = I3DUtil.indexToObject(self.components, getXMLString(xmlFile, key .. "#node"), self.i3dMappings)
        numSlots = math.abs(getXMLInt(xmlFile, key .. "#numSlots") or 0)
    end

    if node ~= nil and numSlots > 0 then
        local place = {}

        place.slots = {}
        place.numUsed = 0
        place.animalType = animalType.type

        local numChildren = getNumOfChildren(node)
        if numSlots > numChildren then
            numSlots = numChildren
            g_logManager:xmlWarning(configFileName, "'numSlots' is greater than available children for '%s', this should be fixed.", key)
        end

        for i = 0, numSlots - 1 do
            local slotNode = getChildAt(node, i)
            table.insert(place.slots, {linkNode = slotNode, loadedMesh = nil, place = place})
        end

        table.insert(livestockTrailerSpec.animalPlaces, place)
        livestockTrailerSpec.animalTypeToPlaces[place.animalType] = place

        spec.addonLoadedChickens = true
    end
end

function LivestockTrailerAddon:loadTransportSounds(xmlFile, key, animalType, configFileName, baseDirectory)
    if hasXMLProperty(xmlFile, key) then
        local spec = self.spec_livestockTrailerAddon

        if spec.animalTypeSamples[animalType.type] == nil then
            local animalTypeSamples = {}

            animalTypeSamples.samples = {}
            animalTypeSamples.numSamples = 0

            animalTypeSamples.minInterval = math.max(Utils.getNoNil(getXMLInt(xmlFile, key .. "#minInterval"), 6000), 3000)
            animalTypeSamples.maxInterval = Utils.getNoNil(getXMLInt(xmlFile, key .. "#maxInterval"), 16000)

            local j = 0
            while true do
                local sample = g_soundManager:loadSampleFromXML(xmlFile, key, string.format("sample(%d)", j), baseDirectory, self.components, 1, AudioGroup.VEHICLE, nil, self)
                if sample == nil or LivestockTrailerAddon.MAXIMUM_ANIMAL_TYPE_SOUNDS < animalTypeSamples.numSamples then
                    break
                end

                animalTypeSamples.numSamples = animalTypeSamples.numSamples + 1
                animalTypeSamples.samples[animalTypeSamples.numSamples] = sample

                j = j + 1
            end

            if animalTypeSamples.numSamples > 0 then
                spec.animalTypeSamples[animalType.type] = animalTypeSamples
            else
                g_logManager:xmlWarning(configFileName, "No samples have been given for '%s', ignoring!", key)
            end
        end
    end
end

function LivestockTrailerAddon:getGlobalSlotLinkNode(slotData)
    local linkNode

    -- Try and get the parent of the first animal if possible as this is the best link node.
    if hasXMLProperty(self.xmlFile, "vehicle.livestockTrailer.animal(0)") then
        linkNode = I3DUtil.indexToObject(self.components, getXMLString(self.xmlFile, "vehicle.livestockTrailer.animal(0)#node"), self.i3dMappings)

        if linkNode ~= nil then
            return getParent(linkNode)
        end
    end

    -- Use backup if no animals exist. If mod has been changed print a warning.
    linkNode = I3DUtil.indexToObject(self.components, slotData.backupLinkNode, self.i3dMappings)

    if linkNode == nil then
        local vehicleData = g_livestockTrailerAddon:getGlobalVehicleData(self.configFileName)
        slotData.canLoad = false

        if vehicleData.isMod then
            g_logManager:warning("[Livestock Trailer Addon]  Failed to add '%s', missing link node for '%s'. Setup was defined for %s (Version: %s). ", animalType.type, self.customEnvironment, vehicleData.version)
        else
            g_logManager:warning("[Livestock Trailer Addon]  Failed to add '%s', missing link node. Setup was defined for game version %s.", animalType.type, vehicleData.version)
        end
    end

    return linkNode
end

function LivestockTrailerAddon:onDelete()
    local spec = self.spec_livestockTrailerAddon

    for _, soundData in pairs (spec.animalTypeSamples) do
        if soundData.samples ~= nil then
            g_soundManager:deleteSamples(soundData.samples)
        end
    end

    for _, placeData in pairs (spec.animalTypePlaces) do
        if placeData.filename ~= nil and placeData.baseDirectory ~= nil then
            g_i3DManager:releaseSharedI3DFile(placeData.filename, placeData.baseDirectory, true)

            placeData.filename = nil
            placeData.baseDirectory = nil
        end
    end
end

function LivestockTrailerAddon:onReadUpdateStream(streamId, timestamp, connection)
    if connection:getIsServer() then
        local spec = self.spec_livestockTrailerAddon

        if streamReadBool(streamId) then
            local sampleIndexOne = streamReadInt8(streamId)
            local sampleIndexTwo = streamReadInt8(streamId)

            local currentAnimalType = self.spec_livestockTrailer.currentAnimalType
            if currentAnimalType ~= nil and spec.animalTypeSamples[currentAnimalType] ~= nil then
                self:playTransportSounds(spec.animalTypeSamples[currentAnimalType], sampleIndexOne, sampleIndexTwo)
            end
        end
    end
end

function LivestockTrailerAddon:onWriteUpdateStream(streamId, connection, dirtyMask)
    if not connection:getIsServer() then
        local spec = self.spec_livestockTrailerAddon

        if streamWriteBool(streamId, bitAND(dirtyMask, spec.dirtyFlag) ~= 0) then
            local sampleIndexOne = spec.sampleIndexOne or 0
            local sampleIndexTwo = spec.sampleIndexTwo or 0

            streamWriteInt8(streamId, sampleIndexOne)
            streamWriteInt8(streamId, sampleIndexTwo)
        end
    end
end

function LivestockTrailerAddon:onUpdate(dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
    if self.isServer and g_livestockTrailerAddon ~= nil then
        local spec = self.spec_livestockTrailerAddon

        if self.spec_livestockTrailer.currentAnimalType ~= nil then
            local currentAnimalType = self.spec_livestockTrailer.currentAnimalType

            if spec.lastAnimalType ~= currentAnimalType then
                self.spec_livestockTrailerAddon.lastAnimalType = currentAnimalType
                spec.soundDelayTime = g_currentMission.time
            end

            local soundData = spec.animalTypeSamples[currentAnimalType]
            if soundData ~= nil then
                if spec.soundDelayTime < g_currentMission.time then
                    local sampleIndexOne = math.random(1, soundData.numSamples)
                    local sampleIndexTwo = sampleIndexOne

                    if (#self.spec_livestockTrailer.loadedAnimals > 1) and (math.random() > 0.5) then
                        sampleIndexTwo = math.random(1, soundData.numSamples)
                    end

                    self:playTransportSounds(soundData, sampleIndexOne, sampleIndexTwo)

                    spec.sampleIndexOne = sampleIndexOne
                    spec.sampleIndexTwo = sampleIndexTwo

                    self:raiseDirtyFlags(spec.dirtyFlag)

                    spec.soundDelayInterval = math.random(soundData.minInterval, soundData.maxInterval)
                    spec.soundDelayTime = g_currentMission.time + spec.soundDelayInterval
                end
            end

            if isActiveForInputIgnoreSelection and VehicleDebug.state == VehicleDebug.DEBUG then
                if not g_gui:getIsGuiVisible() then
                    local sampleIndexOne = spec.sampleIndexOne
                    local sampleIndexTwo = 0

                    if spec.sampleIndexTwo ~= sampleIndexOne then
                        sampleIndexTwo = spec.sampleIndexTwo
                    end

                    local values = {
                        {name = "Vehicle", value = self:getName()},
                        {name = "Animal Type", value = currentAnimalType},
                        {name = "Sample Delay", value = string.format("%02.f seconds", spec.soundDelayInterval / 1000)},
                        {name = "Next Sample In", value = string.format("%02.f seconds", (spec.soundDelayTime - g_currentMission.time) / 1000)},
                        {name = "Sample 1 Index", value = string.format("%01.f", sampleIndexOne)},
                        {name = "Sample 2 Index", value = string.format("%01.f", sampleIndexTwo)},
                        {name = "Play Transport Sounds", value = tostring(g_livestockTrailerAddon.playTransportSounds)}
                    }

                    DebugUtil.renderTable(0.75, 0.55, 0.02, values, 0)
                end
            end

            self:raiseActive()
        end
    end
end

function LivestockTrailerAddon:playTransportSounds(soundData, sampleIndexOne, sampleIndexTwo)
    if self:getCanPlayTransportSounds() then
        local sampleOne = soundData.samples[sampleIndexOne]

        if sampleOne ~= nil then
            local sampleTwo = soundData.samples[sampleIndexTwo]

            g_soundManager:playSample(sampleOne, 0)

            if sampleTwo ~= nil and sampleTwo ~= sampleOne then
                g_soundManager:playSample(sampleTwo, 0, sampleOne)
            end
        end
    end
end

function LivestockTrailerAddon:getCanPlayTransportSounds()
    if g_livestockTrailerAddon ~= nil and g_livestockTrailerAddon.playTransportSounds then
        return not g_gui:getIsGuiVisible()
    end

    return false
end

function LivestockTrailerAddon:getValidDummyFilename(animal)
    if animal.subType.type == "CHICKEN" then
        if self.spec_livestockTrailerAddon.addonLoadedChickens then
            if g_livestockTrailerAddon ~= nil then
                return g_livestockTrailerAddon:getValidDummyFilename(animal.subType.fillType), true
            end
        else
            -- If mod is using LivestockTrailer spec to add chickens and map has default dummy files make sure to use empty 'TG' to stop errors.
            local filename = animal.subType.dummy.filename
            if LivestockTrailerAddon.INVALID_DUMMY_NAMES[filename] ~= nil then
                return filename, false
            end
        end
    end

    return animal.subType.dummy.filename, true
end

function LivestockTrailerAddon:getFillLevelInformation(superFunc, fillLevelInformations)
    superFunc(self, fillLevelInformations)

    if g_livestockTrailerAddon ~= nil and g_livestockTrailerAddon.showAnimalFillLevel then
        local livestockTrailerSpec = self.spec_livestockTrailer
        local animalType = livestockTrailerSpec.currentAnimalType

        if animalType ~= nil then
            local place = livestockTrailerSpec.animalTypeToPlaces[animalType]
            local fillType = g_livestockTrailerAddon:getFirstFillTypeFromAnimalType(animalType)

            if place ~= nil and fillType ~= nil then
                local capacity = #place.slots
                local fillLevel = place.numUsed

                if fillLevel > 0 then
                    local added = false

                    for _, fillLevelInformation in pairs(fillLevelInformations) do
                        if fillLevelInformation.fillType == fillType then
                            fillLevelInformation.fillLevel = fillLevelInformation.fillLevel + fillLevel
                            fillLevelInformation.capacity = fillLevelInformation.capacity + capacity

                            added = true
                            break
                        end
                    end

                    if not added then
                        table.insert(fillLevelInformations, {fillType = fillType, fillLevel = fillLevel, capacity = capacity})
                    end
                end
            end
        end
    end
end

function LivestockTrailerAddon:addAnimal(superFunc, animal)
    local livestockTrailerSpec = self.spec_livestockTrailer
    local success, _ = ListUtil.addElementToList(livestockTrailerSpec.loadedAnimals, animal)

    if success then
        livestockTrailerSpec.currentAnimalType = animal:getSubType().type
        livestockTrailerSpec.loadedAnimalsIds[NetworkUtil.getObjectId(animal)] = true
        livestockTrailerSpec.animalToSlots[animal] = {}

        local place = livestockTrailerSpec.animalTypeToPlaces[animal.subType.type]

        for _, slot in ipairs(place.slots) do
            if slot.loadedMesh == nil then
                local animalRoot
                local filename, isValidDummy = self:getValidDummyFilename(animal)

                if isValidDummy then
                    local animalMeshRoot = g_i3DManager:loadSharedI3DFile(filename)
                    animalRoot = getChildAt(animalMeshRoot, 0)
                    local animalMesh = I3DUtil.indexToObject(animalMeshRoot, animal.subType.dummy.meshNodeStr)
                    local animalHairMesh = I3DUtil.indexToObject(animalMeshRoot, animal.subType.dummy.hairNodeStr)

                    link(slot.linkNode, animalRoot)
                    delete(animalMeshRoot)

                    local x, y, z, w = getShaderParameter(animalMesh, "RDT")
                    local dirt = animal:getDirtScale()

                    setShaderParameter(animalMesh, "RDT", x, dirt, z, w, false)

                    local x, y, _, _ = getShaderParameter(animalMesh, "atlasInvSizeAndOffsetUV")
                    local numTilesU = 1 / x
                    local numTilesV = 1 / y
                    local subType = animal:getSubType()
                    local tileUIndex = subType.texture.tileUIndex
                    local tileVIndex = subType.texture.tileVIndex
                    local tileU = tileUIndex / numTilesU
                    local tileV = tileVIndex / numTilesV

                    setShaderParameter(animalMesh, "atlasInvSizeAndOffsetUV", x, y, tileU, tileV, false)

                    if animalHairMesh ~= nil then
                        local x, y, _, _ = getShaderParameter(animalHairMesh, "atlasInvSizeAndOffsetUV")
                        setShaderParameter(animalHairMesh, "atlasInvSizeAndOffsetUV", x, y, tileU, tileV, false)
                    end
                else
                    animalRoot = createTransformGroup("FakeChickenTG")
                    link(slot.linkNode, animalRoot)
                end

                slot.loadedMesh = animalRoot

                table.insert(livestockTrailerSpec.animalToSlots[animal], slot)

                place.numUsed = place.numUsed + 1

                break
            end
        end

        self:setMassDirty()

        if self.isServer then
            self:raiseDirtyFlags(livestockTrailerSpec.dirtyFlag)
        end
    end
end

function LivestockTrailerAddon.loadSpecValueNumberAnimalsChicken(xmlFile, customEnvironment)
    -- Ignore all Giants vehicles as they will not have the XML property anyway.
    if customEnvironment == nil then
        return
    end

    if hasXMLProperty(xmlFile, "vehicle.livestockTrailerAddon") then
        local i = 0
        while true do
            local key = string.format("vehicle.livestockTrailerAddon.animal(%d)", i)
            if not hasXMLProperty(xmlFile, key) then
                break
            end

            local animalTypeName = getXMLString(xmlFile, key .. "#type")
            if animalTypeName ~= nil and animalTypeName:upper() == "CHICKEN" then
                if g_livestockTrailerAddon ~= nil then
                    g_livestockTrailerAddon:setInvalidGlobalVehicle(customEnvironment, {modName = customEnvironment, reason = "Loaded from vehicle XML."})
                end

                return getXMLInt(xmlFile, key .. "#numSlots")
            end

            i = i + 1
        end
    elseif hasXMLProperty(xmlFile, "vehicle.livestockTrailer") then
        local i = 0
        while true do
            local key = string.format("vehicle.livestockTrailer.animal(%d)", i)
            if not hasXMLProperty(xmlFile, key) then
                break
            end

            local animalTypeName = getXMLString(xmlFile, key .. "#type")
            if animalTypeName ~= nil and animalTypeName:upper() == "CHICKEN" then
                if g_livestockTrailerAddon ~= nil then
                    g_livestockTrailerAddon:setInvalidGlobalVehicle(customEnvironment, {modName = customEnvironment, reason = "Possible incompatibility."})
                end

                return
            end

            i = i + 1
        end
    end

    return
end

function LivestockTrailerAddon.getSpecValueNumberAnimalsChicken(storeItem, realItem)
    return LivestockTrailer.getSpecValueNumberAnimals(storeItem, realItem, "numAnimalsChicken")
end
